###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Pierre-Alain Chaumeil, Aaron Mussig and Donovan Parks'
__author_email__ = 'p.chaumeil@uq.edu.au'
__copyright__ = 'Copyright 2017'
__credits__ = ['Pierre Chaumeil']
__description__ = 'A toolkit for assigning objective taxonomic classifications to bacterial and archaeal genomes.'
__email__ = 'uqpchaum@uq.edu.au'
__license__ = 'GPL3'
__maintainer__ = 'Pierre-Alain Chaumeil, Aaron Mussig and Donovan Parks'
__maintainer_email__ = 'donovan.parks@gmail.com'
__name__ = 'gtdbtk'
__python_requires__ = '>=3.6'
__status__ = 'Production'
__title__ = 'GTDB-Tk'
__url__ = 'https://github.com/Ecogenomics/GTDBTk'
__version__ = '1.7.0'
